
# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network models and simulations. See also: 
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
""" Tests that go all the way from YAML-based templates to the backend representation and potentially computations."""

__author__ = "Daniel Rose"
__status__ = "Development"

import pytest


def setup_module():
    print("\n")
    print("========================================")
    print("| Test Suite: From Frontend to Backend |")
    print("========================================")


@pytest.mark.xfail
def test_simple_example():
    """Test of a simple self-connecting one-node backend with a linear operator for the full pipeline from YAML
    to simulation."""

    # Step 1: Load Circuit template
    from pyrates.frontend.template.circuit import CircuitTemplate
    path = "../model_templates/test_resources/linear/ExampleCircuit"
    tmp = CircuitTemplate.from_yaml(path)

    # Step 2: Instantiate template to create frontend IR
    circuit = tmp.apply()

    # Step 3: Reformat frontend IR to backend IR
    # ToDo: adapt this step to new frontend-ir-backend structure
    from pyrates.frontend.nxgraph import from_circuit
    net_def = from_circuit(circuit, revert_node_names=True)

    # Step 4: Create tensorflow graph
    from pyrates.backend import ComputeGraph
    net = ComputeGraph(net_def, dt=5e-4, vectorization='none')

    # Step 5: Run simulation
    results, _ = net.run(simulation_time=1.,
                         outputs={"V": ('Node.0', 'LinearOperator.0', 'y')},
                         sampling_step_size=1e-3)

    results.plot()


@pytest.mark.xfail
@pytest.mark.parametrize("vectorize", ["none", "nodes", "ops"])
def test_3_coupled_jansen_rit_circuits(vectorize):
    """Test the simple Jansen-Rit example with three coupled circuits."""

    # Step 1: Load Circuit template
    from pyrates.frontend.template.circuit import CircuitTemplate
    path = "model_templates.jansen_rit.circuit.MultiJansenRitCircuit"
    tmp = CircuitTemplate.from_yaml(path)

    # Step 2: Instantiate template to create frontend IR
    circuit = tmp.apply()

    # Step 3: Reformat frontend IR to backend IR
    # ToDo: adapt this step to new frontend-ir-backend structure
    from pyrates.frontend.nxgraph import from_circuit
    net_def = from_circuit(circuit, revert_node_names=True)

    # Step 4: Create tensorflow graph
    from pyrates.backend import ComputeGraph
    net = ComputeGraph(net_def, dt=5e-4, vectorization=vectorize)



"""Test suit for saving PyRates objects to or loading them from file."""

__author__ = "Daniel Rose"
__status__ = "Development"

import filecmp
import os
import pathlib

import pytest


def setup_module():
    print("\n")
    print("=======================")
    print("| Test Suite File I/O |")
    print("=======================")


class FileCompareError(Exception):
    pass


def compare_files(filename1, filename2):
    """Helper function to compare files"""

    if os.path.getsize(filename1) == os.path.getsize(filename2):
        if filecmp.cmp(filename1, filename2, shallow=False):
            pass
        else:
            raise FileCompareError("Files have different content.")
    else:
        raise FileCompareError("Files are not of the same size.")


def get_parent_directory():
    """Helper function to get path of current file"""
    return pathlib.Path(__file__).parent.absolute()


def test_pickle_ir():
    pass

    path = "model_templates.jansen_rit.circuit.JansenRitCircuit"
    from pyrates.frontend.template.circuit import CircuitTemplate
    from pyrates.ir.circuit import CircuitIR
    from pyrates.frontend.fileio import pickle

    template = CircuitTemplate.from_yaml(path)

    # try to pickle non-vectorized circuit
    circuit = template.apply()  # type: CircuitIR
    filename1 = os.path.join(get_parent_directory(), "output", "jansen_rit_ir.p")
    circuit.to_file(filename=filename1, filetype="pickle")

    # compare to reference pickle
    # compare_files("output/jansen_rit_ir.p", "resources/jansen_rit_ir.p")  # currently does not work
    circuit2 = CircuitIR.from_file(filename1, filetype="pickle")

    # ToDo: compare circuit IR instances at runtime

    # try to pickle vectorized circuit
    circuit.optimize_graph_in_place()
    filename2 = os.path.join(get_parent_directory(), "output", "jansen_rit_ir_vectorized.p")
    circuit.to_file(filename=filename2, filetype="pickle")

    # compare to reference pickle
    # compare_files("output/jansen_rit_ir_vectorized.p", "resources/jansen_rit_ir_vectorized.p")
    # currently does not work

    circuit3 = CircuitIR.from_file(filename2, filetype="pickle")

    # ToDo: compare vectorized circuit IR instances at runtime


def test_pickle_template():
    path = "model_templates.jansen_rit.circuit.JansenRitCircuit"
    from pyrates.frontend.template import from_yaml, clear_cache
    clear_cache()
    template = from_yaml(path)
    # include pickle protocol number in file name for version interoperability
    from pickle import HIGHEST_PROTOCOL
    ext = f"p{HIGHEST_PROTOCOL}"
    # configure filenames
    out_file = os.path.join(get_parent_directory(), "output", f"jansen_rit_template.{ext}")
    test_file = os.path.join(get_parent_directory(), "resources", f"jansen_rit_template.{ext}")

    from pyrates.frontend.fileio import pickle

    pickle.dump(template, out_file)
    # to update the reference dump, uncomment the following
    # pickle.dump(template, test_file)

    compare_files(out_file, test_file)

    template = pickle.load(out_file)
    assert template



"""Test suite for accurate behavior of models provided under pyrates.model_templates"""

# external imports
from typing import Union

import numpy as np
import pytest

# pyrates internal imports
from pyrates.ir import CircuitIR

# meta infos
__author__ = "Richard Gast"
__status__ = "Development"


###########
# Utility #
###########


def setup_module():
    print("\n")
    print("================================")
    print("| Test Suite 3 : SimNet Module |")
    print("================================")


def nmrse(x: np.ndarray,
          y: np.ndarray
          ) -> Union[float, np.ndarray]:
    """Calculates the normalized root mean squared error of two vectors of equal length.
    Parameters
    ----------
    x,y
        Arrays to calculate the nmrse between.
    Returns
    -------
    float
        Normalized root mean squared error.
    """

    max_val = np.max((np.max(x, axis=0), np.max(y, axis=0)))
    min_val = np.min((np.min(x, axis=0), np.min(y, axis=0)))

    diff = x - y

    return np.sqrt(np.sum(diff ** 2, axis=0)) / (max_val - min_val)


#########
# Tests #
#########


def test_3_1_jansenrit():
    """Testing accuracy of Jansen-Rit model implementations:
    """

    T = 1.5
    dt = 1e-4
    dts = 1e-2

    # compare single operator JRC implementation with multi-node JRC implementation
    ###############################################################################

    # single operator JRC
    jrc1 = CircuitIR.from_yaml("model_templates.jansen_rit.simple_jansenrit.JRC_simple")
    jrc1 = jrc1.compile(backend='numpy', step_size=dt, solver='scipy')
    r1 = jrc1.run(T, outputs={'EIN': 'JRC/JRC_op/PSP_ein'}, sampling_step_size=dts)
    jrc1.clear()

    # multi-node JRC
    jrc2 = CircuitIR.from_yaml("model_templates.jansen_rit.simple_jansenrit.JRC")
    jrc2 = jrc2.compile(backend='numpy', step_size=dt, solver='scipy')
    r2 = jrc2.run(T, outputs={'EIN': 'EIN/RPO_e/PSP'}, sampling_step_size=dts)
    jrc2.clear()

    assert np.mean(r1.values.flatten() - r2.values.flatten()) == pytest.approx(0., rel=1e-4, abs=1e-4)


def test_3_2_montbrio():
    """Testing accuracy of mean-field representation of QIF population.
    """

    T = 80.0
    dt = 1e-3
    dts = 1e-2

    in_start = int(np.round(30.0/dt))
    in_dur = int(np.round(20.0/dt))
    inp = np.zeros((int(np.round(T/dt)),))
    inp[in_start:in_start+in_dur] = 5.0

    # assess correct behavior of the model around the bi-stable regime
    ##################################################################

    # set up circuit
    m1 = CircuitIR.from_yaml("model_templates.montbrio.simple_montbrio.QIF_exc")
    m1 = m1.compile(vectorization=True, backend='numpy', solver='scipy', step_size=dt)

    # perform simulation
    r1 = m1.run(T, sampling_step_size=dts, inputs={"p/Op_e/inp": inp}, outputs={"r": "p/Op_e/r"})
    m1.clear()

    # test firing rate relationships at pre-defined times
    times = [25.0, 49.0, 79.0]
    indices = [np.argmin(np.abs(t - r1.index)) for t in times]
    assert r1.iloc[indices[0], 0] < r1.iloc[indices[1], 0]
    assert r1.iloc[indices[0], 0] < r1.iloc[indices[2], 0]
    assert r1.iloc[indices[1], 0] > r1.iloc[indices[2], 0]


def test_3_3_wilson_cowan():
    """Test accuracy of wilson cowan neural mass model implementation.
    """

    # assess correct response of model to input
    ###########################################

    T = 80.0
    dt = 1e-3
    dts = 1e-2

    in_start = int(np.round(30.0 / dt))
    in_dur = int(np.round(20.0 / dt))
    inp = np.zeros((int(np.round(T / dt)),))
    inp[in_start:in_start + in_dur] = 5.0

    # set up circuit
    wc1 = CircuitIR.from_yaml("model_templates.wilson_cowan.simple_wilsoncowan.WC_simple")
    wc1 = wc1.compile(vectorization=True, backend='numpy', solver='scipy', step_size=dt)

    # perform simulation
    r1 = wc1.run(T, sampling_step_size=dts, inputs={"E/Op_rate/I_ext": inp}, outputs={"R_e": "E/Op_rate/r"})
    wc1.clear()

    # test firing rate relationships at pre-defined times
    times = [29.0, 49.0, 79.0]
    indices = [np.argmin(np.abs(t - r1.index)) for t in times]
    assert r1.iloc[indices[0], 0] < r1.iloc[indices[1], 0]
    assert r1.iloc[indices[0], 0] - r1.iloc[indices[2], 0] == pytest.approx(0., rel=1e-4, abs=1e-4)

    # repeat for model with short-term plasticity
    #############################################

    T = 220.0
    dt = 5e-3
    dts = 1e-2

    in_start = int(np.round(30.0 / dt))
    in_dur = int(np.round(20.0 / dt))
    inp = np.zeros((int(np.round(T / dt)),))
    inp[in_start:in_start + in_dur] = 5.0

    # set up circuit
    wc2 = CircuitIR.from_yaml("model_templates.wilson_cowan.simple_wilsoncowan.WC_stp")
    wc2 = wc2.compile(vectorization=True, backend='numpy', solver='scipy', step_size=dt)

    # perform simulation
    r2 = wc2.run(T, sampling_step_size=dts, inputs={"E/E_op/I_ext": inp}, outputs={"V_e": "E/E_op/v"})
    wc2.clear()

    # test firing rate relationships at pre-defined times
    times = [29.0, 49.0, 219.0]
    indices = [np.argmin(np.abs(t - r2.index)) for t in times]
    assert r2.iloc[indices[0], 0] < r2.iloc[indices[1], 0]
    assert r2.iloc[indices[0], 0] - r2.iloc[indices[2], 0] == pytest.approx(0., rel=1e-3, abs=1e-3)


def test_3_4_kuramoto():
    """Tests accurate behavior of kuramoto oscillator model.
    """

    # assess correct response of single base oscillator
    ###################################################

    T = 2.0
    dt = 1e-4
    dts = 1e-2

    # set up circuit
    km1 = CircuitIR.from_yaml("model_templates.kuramoto.simple_kuramoto.KM_single")
    km1 = km1.compile(vectorization=True, backend='numpy', solver='scipy', step_size=dt)

    # perform simulation
    r1 = km1.run(T, sampling_step_size=dts, outputs={"theta": "p1/Op_base/theta"})
    km1.clear()

    # test linear oscillator properties
    omega = 10.0
    results = np.sin(r1.values[:, 0]*2*np.pi)
    target = np.sin(omega*2.0*np.pi*r1.index.values)
    assert results - target == pytest.approx(0., rel=1e-2, abs=1e-2)

    # assess correct response of two coupled oscillators
    ####################################################

    T = 6.0
    dt = 1e-4
    dts = 1e-2

    # set up circuit
    km2 = CircuitIR.from_yaml("model_templates.kuramoto.simple_kuramoto.KMN")
    km2 = km2.compile(vectorization=True, backend='numpy', solver='scipy', step_size=dt)

    # perform simulation
    r2 = km2.run(T, sampling_step_size=dts, outputs={"theta1": "p1/Op_base/theta", "theta2": "p2/Op_base/theta"})
    km2.clear()

    # test whether oscillator 2 showed a faster phase development than oscillator 1
    assert r2['theta1'].iloc[-1, 0] < r2['theta2'].iloc[-1, 0]

    # repeat test 2 for two coupled noisy oscillators
    #################################################

    T = 6.0
    dt = 1e-4
    dts = 1e-2

    inp1 = np.random.randn(int(np.round(T / dt))) * 0.5
    inp2 = np.random.randn(int(np.round(T / dt))) * 0.1

    # set up circuit
    km3 = CircuitIR.from_yaml("model_templates.kuramoto.simple_kuramoto.KMN_noise")
    km3 = km3.compile(vectorization=True, backend='numpy', solver='scipy', step_size=dt)

    # perform simulation
    r3 = km3.run(T, sampling_step_size=dts, outputs={"theta1": "p1/Op_noise/theta", "theta2": "p2/Op_noise/theta"},
                 inputs={"p1/Op_noise/xi": inp1, "p2/Op_noise/xi": inp2})
    km3.clear()

    # test whether oscillator 2 showed a faster phase development than oscillator 1
    assert r3['theta1'].iloc[-1, 0] < r3['theta2'].iloc[-1, 0]




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network models and simulations. See also: 
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
""" Tests for the parser that translates circuits and components defined in YAML into the intermediate
python representation.
"""

__author__ = "Daniel Rose"
__status__ = "Development"

import pytest


def setup_module():
    print("\n")
    print("===========================")
    print("| Test Suite: YAML Parser |")
    print("===========================")


@pytest.mark.parametrize("operator", ["model_templates.jansen_rit.axon.axon.PotentialToRateOperator",
                                      "model_templates.jansen_rit.axon.templates.SigmoidPRO",
                                      "model_templates.jansen_rit.axon.templates.JansenRitPRO",
                                      "model_templates.jansen_rit.population.population.CurrentToPotentialOperator",
                                      "model_templates.jansen_rit.synapse.synapse.RateToCurrentOperator"
                                      ])
def test_import_operator_templates(operator):
    """test basic (vanilla) YAML parsing using ruamel.yaml (for YAML 1.2 support)"""
    from pyrates.frontend.template.operator import OperatorTemplate
    from pyrates.frontend.template import template_cache, clear_cache
    clear_cache()

    template = OperatorTemplate.from_yaml(operator)  # type: OperatorTemplate

    assert template.path in template_cache

    cached_template = template_cache[operator]  # type: OperatorTemplate
    assert template is cached_template
    assert template.path == cached_template.path
    assert template.equations == cached_template.equations
    assert repr(template) == repr(cached_template) == f"<OperatorTemplate '{operator}'>"


# @pytest.mark.parametrize("node", ["pyrates.population.templates.JansenRitIN",
#                                   "pyrates.population.templates.JansenRitPC",
#                                   "pyrates.population.population.NeuralMass"
#                                   ])
# def test_import_node_templates(node):
#     """test import of node templates"""
#
#     from pyrates.utility.yaml_parser import TemplateLoader
#     from pyrates.node.node import NodeTemplateLoader
#
#     template = NodeTemplate.from_yaml(node)  # type: NodeTemplate
#
#     assert template.path in TemplateLoader.cache  # just to check if cache is really shared among subclasses
#
#     cached_template = NodeTemplateLoader.cache[node]
#     assert template is cached_template
#     assert template.path == cached_template.path
#     for op in template.operators:
#         assert isinstance(op, OperatorTemplate)
#     assert template.operators == cached_template.operators
#     assert repr(template) == repr(cached_template) == f"<NodeTemplate '{node}'>"


def test_full_jansen_rit_circuit_template_load():
    """Test a simple circuit template, including all nodes and operators to be loaded."""

    path = "model_templates.jansen_rit.circuit.JansenRitCircuit"
    from pyrates.frontend.template.circuit import CircuitTemplate
    from pyrates.frontend.template.edge import EdgeTemplate
    from pyrates.frontend.template.node import NodeTemplate
    from pyrates.frontend.template.operator import OperatorTemplate
    from pyrates.frontend.template import template_cache, clear_cache
    clear_cache()

    template = CircuitTemplate.from_yaml(path)

    # test, whether circuit is in loader cache
    assert template is template_cache[path]

    # test, whether node templates have been loaded successfully
    nodes = {"JR_PC": "model_templates.jansen_rit.population.templates.JansenRitPC",
             "JR_IIN": "model_templates.jansen_rit.population.templates.JansenRitIN",
             "JR_EIN": "model_templates.jansen_rit.population.templates.JansenRitIN"}

    for key, value in nodes.items():
        assert isinstance(template.nodes[key], NodeTemplate)
        assert template.nodes[key] is template_cache[value]
        # test operators in node templates
        for op in template.nodes[key].operators:
            assert op.path in template_cache
            assert isinstance(op, OperatorTemplate)

    # test, whether coupling operator has been loaded correctly
    coupling_path = "model_templates.jansen_rit.edges.LinearCouplingOperator"
    edge_temp = template.edges[0][2]
    assert isinstance(edge_temp, EdgeTemplate)
    assert list(edge_temp.operators)[0] is template_cache[coupling_path]

    assert repr(template) == f"<CircuitTemplate '{path}'>"


def test_circuit_instantiation():
    """Test, if apply() functions all work properly"""
    path = "model_templates.jansen_rit.circuit.JansenRitCircuit"
    from pyrates.frontend import template
    template.clear_cache()

    circuit_template = template.from_yaml(path)

    circuit = circuit_template.apply()
    # used to be: test if two edges refer to the same coupling operator by comparing ids
    # this is why we referenced by "operator"
    # now: compare operators directly
    edge_to_compare = circuit.edges[('JR_PC', 'JR_EIN', 0)]["edge_ir"]
    for op_key, op in circuit.edges[("JR_PC", "JR_IIN", 0)]["edge_ir"].op_graph.nodes(data=True):
        if op_key in edge_to_compare:
            assert op["operator"].equations == edge_to_compare[op_key].equations

    # now test, if JR_EIN and JR_IIN refer to the same operator graph
    assert circuit["JR_EIN"].op_graph is circuit["JR_IIN"].op_graph

    # now test, if the references are collected properly
    for key, data in circuit.nodes(data=True):
        node = data["node"]
        assert node in circuit._reference_map[node.op_graph]

    assert len(circuit._reference_map[circuit["JR_EIN"].op_graph]) == 2

    # verify that .apply also understands value updates to nodes
    value_dict = {"JR_PC/JansenRitExcitatorySynapseRCO/h": 0.1234}
    circuit2 = circuit_template.apply(node_values=value_dict)
    node = circuit2["JR_PC"]
    assert node.values["JansenRitExcitatorySynapseRCO"]["h"] == 0.1234


def test_multi_circuit_instantiation():
    """Test, if a circuit with subcircuits is also working."""
    path = "model_templates.jansen_rit.circuit.MultiJansenRitCircuit"
    from pyrates.frontend import template as tpl
    tpl.clear_cache()

    template = tpl.from_yaml(path)

    circuit = template.apply()
    assert circuit


def test_equation_alteration():
    """Test, if properties of a template that mean to alter a certain parent equation are treated correctly"""

    path = "model_templates.jansen_rit.population.templates.InstantaneousCPO"
    # this template removes the component "L_m * " from the base equation "L_m * V = k * I"
    from pyrates.frontend.template.operator import OperatorTemplate

    template = OperatorTemplate.from_yaml(path)

    operator, values = template.apply()

    assert operator.equations == ("V = k * I",)


def test_yaml_dump():
    """Test the functionality to dump an object to YAML"""
    from pyrates.frontend import fileio

    with pytest.raises(AttributeError):
        fileio.dump("no_to_dict()", "random_art", "yaml")

    from pyrates.frontend.template.circuit import CircuitTemplate
    circuit = CircuitTemplate.from_yaml("model_templates.jansen_rit.circuit.JansenRitCircuit").apply()

    with pytest.raises(ValueError):
        fileio.dump(circuit, "output/yaml_dump.yaml", "yml")

    with pytest.raises(TypeError):
        fileio.dump(circuit, "output/yaml_dump.yaml", "yaml")


    fileio.dump(circuit, "output/yaml_dump.yaml", "yaml", "DumpedCircuit")

    # reload saved circuit
    saved_circuit = CircuitTemplate.from_yaml("output/yaml_dump/DumpedCircuit").apply()
    assert saved_circuit

    # ToDo: figure out a simple way to compare both instances



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

# external imports
import numpy as np
import pandas as pd
from typing import Optional, Union
from scipy.optimize import differential_evolution

# system imports
import os
import sys
import glob
import time as t
from itertools import cycle

# pyrates imports
from pyrates.utility.grid_search import grid_search, ClusterGridSearch, linearize_grid, adapt_circuit, ClusterCompute, \
    StreamTee
from pyrates.frontend import CircuitTemplate

# meta infos
__author__ = "Christoph Salomon, Richard Gast"
__status__ = "development"


class GeneticAlgorithmTemplate:
    def __init__(self):

        # Initialize storage variables
        self.initial_gene_pool = {}
        self.num_genes = 0
        self.sigma_adapt = 0
        self.gene_names = []
        self.pop = pd.DataFrame()
        self.pop_size = 0
        self.candidate = pd.DataFrame()
        self.winner = pd.DataFrame()
        self.current_winners = pd.DataFrame()
        self.current_max_fitness = 0
        self.drop_count = 0

    def run(self, initial_gene_pool: dict, target: list, max_iter: int, min_fit: Optional[float] = 0.,
            n_winners: Optional[int] = 1, n_parent_pairs: Optional[int] = 10, n_new: Optional[int] = 0,
            sigma_adapt: Optional[float] = 0., max_stagnation_steps: Optional[int] = 0,
            stagnation_decimals: Optional[int] = 8, max_stagnation_drops: Optional[Union[int, float]] = np.Inf,
            enforce_max_iter: Optional[bool] = False, new_pop_on_drop: Optional[bool] = False,
            pop_save: Optional[str] = "", candidate_save: Optional[str] = "", drop_save: Optional[str] = "",
            gene_sampling_func=np.linspace, new_member_sampling_func=None, permute: bool = True, **kwargs):
        """Run a genetic algorithm to optimize genes of a population in respect to a given target vector

        Parameters
        ----------
        initial_gene_pool
            Dictionary containing ranges for each gene to sample from
        target
            Target values that are used to determine the fitness of a population member
        max_iter
            Maximum number of iterations (generations) before the computation is terminated
        min_fit
            Minimum fitness. If set, computation will stop after the fitness of one population member exceeds this value
        n_winners
            Number of strongest members of a population, that will be members of this population's offspring
        n_parent_pairs
            Number of parent pairs of a population that will produce a child which will be member of this population's
            offspring (crossover)
        n_new
            Number of members in a population's offspring, that will be created from the initial gene pool
        sigma_adapt
            Ratio by which the mean derivation of the gene probability distributions will change during a mutation as
            suggested by (Beyer1995 - Toward a Theory of Evolution Strategies: Self-Adaptation)
        max_stagnation_steps
            Maximum number of iterations with no change in the fitness, before the strongest member of a population
            will be discarded and not be part of this population's offspring anymore
        stagnation_decimals
            Decimal precision that will be used to detect changes in the fitness of successive populations
        max_stagnation_drops
            If True, computation will stop when the maximum stagnation is reached
        enforce_max_iter
            If True, all iterations will performed, even if another convergence criterion is reached before
        new_pop_on_drop
            If True, a new population is created once the fitness stagnates. If False, only the fittest member of
            the population is replaced by a new member and the computation continues
        pop_save
            If set, the whole population will be saved to an hdf5 file.
        candidate_save
            If set, the strongest member of a population will be saved to an hdf5 file, before the population is updated
        drop_save
            If set, all members that are dropped from a population due to stagnation or other criteria will be saved to
            this folder in hdf5 format
        permute
            If False, no permutations of gene samples will be performed. Instead, the gene sampling size equals the
            population size.
        gene_sampling_func
        new_member_sampling_func
        kwargs


        Returns
        -------
        pandas.DataFrame containing the overall fittest member of all computed populations

        """

        import h5py

        self.initial_gene_pool = initial_gene_pool
        self.num_genes = len(initial_gene_pool)
        self.sigma_adapt = sigma_adapt

        # Counts how many members have already been dropped out from a population due to fitness stagnation
        self.drop_count = 0

        # Create starting population
        ############################
        self.__create_pop(sampling_func=gene_sampling_func, permute=permute)
        self.pop_size = self.pop.shape[0]

        if n_parent_pairs + n_new > self.pop_size or n_winners > self.pop_size:
            print('WARNING: Number of winners, or parents and new members exceeds the population size. Returning')
            return

        # Start genetic algorithm
        #########################
        print("***STARTING GENETIC ALGORITHM***")
        iter_count = 0
        stagnation_count = 0
        while iter_count < max_iter:

            print("")
            print(f'ITERATION: {iter_count}')

            # Evaluate fitness of current population
            ########################################
            self.eval_fitness(target, **kwargs)
            new_candidate = self.pop.nlargest(1, "fitness")
            self.current_max_fitness = float(new_candidate.loc[:, "fitness"].values)

            # If no population member yields a proper fitness value since all computed timeseries contained at least one
            # undefined value (e.g. np.NaN)
            print(f'Fittest gene in current population:')
            self.plot_genes(new_candidate)
            target_tmp = []
            for tar in target:
                if isinstance(tar, list):
                    target_tmp.append(np.round(tar, decimals=2))
                else:
                    target_tmp.append(np.round(tar, decimals=2))
            print(f'Target: {target_tmp}')

            # Check for fitness stagnation
            ##############################
            if max_stagnation_steps > 0:
                # Before the first iteration self.candidate is empty. Skip stagnation check in that case
                if not self.candidate.empty:
                    old_fitness = np.round(float(self.candidate['fitness']), decimals=stagnation_decimals)
                    new_fitness = np.round(self.current_max_fitness, decimals=stagnation_decimals)
                    # Check for change in fitness
                    if new_fitness <= old_fitness:
                        stagnation_count += 1
                        # Check if stagnation occured
                        if stagnation_count > max_stagnation_steps:
                            print("Maximum fitness stagnation reached!")
                            # Check if maximum number of population drops is reached
                            if not (self.drop_count == max_stagnation_drops) or enforce_max_iter:
                                if drop_save:
                                    print("Saving fittest candidate.")
                                    os.makedirs(drop_save, exist_ok=True)
                                    self.candidate.to_hdf(f'{drop_save}/PopulationDrop_{self.drop_count}.h5',
                                                          key='data')
                                    with h5py.File(f'{drop_save}/PopulationDrop_{self.drop_count}.h5') as file:
                                        file['target'] = target

                                self.drop_count += 1
                                self.candidate = pd.DataFrame()

                                if new_pop_on_drop:
                                    print("Creating new population.")
                                    self.__create_pop(sampling_func=gene_sampling_func, permute=permute)
                                    continue
                                else:
                                    print("Dropping candidate from population!")
                                    self.current_winners = self.current_winners.drop(self.candidate.index)
                            else:
                                print("Returning fittest member!")
                                print("")
                                return self.winner
                    else:
                        # Reset stagnation counter
                        stagnation_count = 0

            # Update candidate and save if necessary
            ########################################
            self.candidate = new_candidate
            if pop_save:
                self.pop.to_hdf(f"{pop_save}_{iter_count}.h5", key='data', mode='w')
            elif candidate_save:
                self.candidate.to_hdf(candidate_save, key='data', mode='w')

            # Update current winning genes
            ##############################
            if self.winner.empty or float(self.candidate['fitness']) > float(self.winner['fitness']):
                self.winner = self.candidate
                print('Fittest gene in current population is also the globally fittest gene.')
            else:
                print(f'Globally fittest gene:')
                self.plot_genes(self.winner)
                target_tmp = []
                for tar in target:
                    if isinstance(tar, list):
                        target_tmp.append(np.round(tar, decimals=2))
                    else:
                        target_tmp.append(np.round(tar, decimals=2))
                print(f'Target: {target_tmp}')

            # Evaluate minimum fitness conversion criteria
            ##############################################
            if 0 < min_fit < self.current_max_fitness:
                print("Minimum fitness criterion reached!")
                if enforce_max_iter:
                    if drop_save:
                        print("Saving winner!")
                        self.winner.to_hdf(f'{drop_save}/PopulationDrop_{self.drop_count}.h5', key='data')
                    self.drop_count += 1
                    stagnation_count = 0
                    self.winner = pd.DataFrame()
                    if new_pop_on_drop:
                        print(f'Generating new population')
                        self.__create_pop(sampling_func=gene_sampling_func, permute=permute)
                        continue
                    else:
                        print("Dropping candidate from population!")
                        self.pop = self.pop.drop(new_candidate.index)
                        self.current_winners = self.current_winners.drop(new_candidate.index)
                        self.candidate = pd.DataFrame()
                else:
                    return new_candidate

            # Create offspring from current population
            ##########################################

            if self.current_max_fitness == -0.0:
                print(f'No candidate available for the current gene set')
                print(f'Generating new population')
                self.__create_pop(sampling_func=gene_sampling_func, permute=permute)
            else:
                print(f'Generating offspring')
                self.__create_offspring(n_parent_pairs=n_parent_pairs, n_new=n_new, n_winners=n_winners,
                                        sampling_func=new_member_sampling_func if new_member_sampling_func
                                        else gene_sampling_func)
            iter_count += 1

        # End of iteration loop
        print("Maximum iterations reached")
        if float(self.winner['fitness']) < min_fit and self.drop_count == 0:
            print('Could not satisfy minimum fitness condition.')
        return self.winner

    def __create_offspring(self, n_winners, n_parent_pairs=0, n_new=0, sampling_func=np.random.uniform):
        """Create a new offspring of the current population

        Offspring contains:
        - n_winners strongest members of the current population (winners)
        - n_parent_pairs children of current parent pairings (crossover)
        - n_mut Mutations of winners and children (mutations)
        - n_new Fresh members based on the initial gene pool (new)

        The number of mutations is chosen dynamically to resize the offspring to the size of the current population
        with n_mut = population_size - n_winners - n_parent_pairs - n_new
        """
        print('Updating population')

        # Create new offspring
        ######################

        offspring = []
        new_sigs = []
        n_mutations = self.pop_size - (n_parent_pairs + n_new)

        # 1. Add n_winners strongest members
        ####################################
        winners = self.__select_winners(n_winners=n_winners)
        winner_genes, winner_sigmas = [], []
        for w in winners:
            winner_genes.append(w[0])
            winner_sigmas.append(w[1])

        # 2. Add children of n_parents parent pairs
        ###########################################
        parent_pairs = self.__create_parent_pairs(n_parent_pairs=n_parent_pairs)
        childs = self.__crossover(parent_pairs)
        for c in childs:
            offspring.append(c[0])
            new_sigs.append(c[1])

        # Each failed child will be replaced by a mutation
        new_mutations = n_parent_pairs - len(childs)
        if new_mutations > 0:
            n_mutations += new_mutations

        # 3. Add mutations
        ##################
        parent_pool = cycle(zip(offspring+winner_genes, new_sigs+winner_sigmas))
        for _ in range(n_mutations):
            parent = next(parent_pool)
            mutation = self.__mutate(parent)
            offspring.append(mutation[0])
            new_sigs.append(mutation[1])

        # 4. Add n_new fresh members from initial gene_pool
        ###################################################
        for n in range(n_new):
            new_member = self.__create_new_member(sampling_func=sampling_func)
            offspring.append(new_member[0])
            new_sigs.append(new_member[1])

        offspring = pd.DataFrame(offspring)

        # 5. Swap possible duplicates in the offspring with new members
        ###############################################################
        while any(offspring.duplicated(keep='first')):
            dupl_idx = offspring.loc[offspring.duplicated(keep='first')].index.to_numpy()
            for i in dupl_idx:
                # Replace every duplicate with a new chromosome, fitness 0.0 and respective sigmas
                new_member = self.__create_new_member(sampling_func=sampling_func)
                offspring.iloc[i] = new_member[0]
                new_sigs[i] = new_member[1]

        offspring.columns = self.pop.loc[:, self.gene_names].columns
        offspring['fitness'] = 0.0
        offspring['sigma'] = new_sigs
        offspring['results'] = [[] for _ in range(len(new_sigs))]
        offspring.columns = self.pop.columns

        self.pop = offspring

    def __create_pop(self, sampling_func=np.linspace, permute=True):
        """Create new population from the initial gene pool"""
        for idx in self.current_winners.index:
            self.current_winners.drop(index=idx)
        pop_grid = {}
        # Prevent duplicates if create_pop() is called again if population had no winner
        self.gene_names = []
        sigmas = [self.initial_gene_pool[gene]['sigma'] for gene in self.initial_gene_pool.keys()]
        for param, value in self.initial_gene_pool.items():
            self.gene_names.append(param)
            value_tmp = value.copy()
            value_tmp.pop('sigma')
            pop_grid[param] = self.__sample_gene(sampling_func, **value_tmp)
        self.pop = linearize_grid(pop_grid, permute=permute)
        self.pop_size = self.pop.shape[0]

        self.pop['fitness'] = 0.0
        self.pop['sigma'] = [sigmas for _ in range(self.pop_size)]
        self.pop['results'] = [[] for _ in range(self.pop_size)]

    def __select_winners(self, n_winners):
        """Returns the n_winners fittest members from the current population"""
        if self.current_winners.shape[0] == n_winners:
            for i in range(n_winners):
                winner = self.pop.nlargest(1, 'fitness').index[0]
                idx_old = self.current_winners.nsmallest(1, 'fitness').index[0]
                if self.pop.at[winner, 'fitness'] > self.current_winners.at[idx_old, 'fitness'] and not \
                        (self.current_winners.loc[:, self.gene_names] == self.pop.loc[winner, self.gene_names]
                         ).all(1).any():
                    self.current_winners.loc[idx_old, :] = self.pop.loc[winner, :]
        else:
            self.current_winners = self.pop.nlargest(n_winners, 'fitness')
        return [(self.current_winners.loc[i, self.gene_names], self.current_winners.at[i, 'sigma'])
                for i in self.current_winners.index]

    def __mutate(self, parent, max_iter=1000):
        """Create mutation of a parent, based on a gaussian distribution for each gene"""
        mu_new = []
        sigma_new = []
        for i, (mu, sigma) in enumerate(zip(parent[0], parent[1])):
            mu_tmp = np.random.normal(mu, sigma)
            sigma_tmp = sigma
            j = 0
            while any([mu_tmp < self.initial_gene_pool[self.gene_names[i]]['min'],
                       mu_tmp > self.initial_gene_pool[self.gene_names[i]]['max']]) and j < max_iter:
                mu_tmp = np.random.normal(mu, sigma_tmp)
                sigma_tmp *= 0.99
                j += 1
            if j > 1:
                sigma = sigma_tmp
            mu_new.append(mu_tmp if j < max_iter else mu)
            # Adapt sigma (Beyer1995, p.5)
            xi = np.exp(self.sigma_adapt*np.random.randn())
            sigma_new.append(sigma*xi)
        return mu_new, sigma_new

    def __create_new_member(self, sampling_func=np.random.uniform):
        """Create a new population member from the initial gene pool"""
        genes = []
        sigma = []
        for i, (key, value) in enumerate(self.initial_gene_pool.items()):
            value_tmp = value.copy()
            value_tmp['size'] = 1
            sigma.append(value_tmp.pop('sigma'))
            genes.append(self.__sample_gene(sampling_func, **value_tmp)[0])
        new_member = [genes, sigma]
        already_exists = (self.pop.loc[:, self.gene_names] == genes).all(1).any()
        if already_exists:
            new_member = self.__create_new_member(sampling_func)
        return new_member

    def __create_parent_pairs(self, n_parent_pairs):
        """Create n_parent_pairs parent combinations. The occurrence probability for each parent is based on that
        parent's fitness"""
        parents = []

        # Reproduction probability for each parent is based on its relative fitness
        pop_fitness = self.pop['fitness'].sum()
        parent_repro = self.current_winners['fitness'].copy()

        # Set -inf and NaN to 0 since np.choice can only handle positive floats or ints
        # Safety measure, should not occur in the first place
        parent_repro[np.isinf(parent_repro)] = 0.0
        parent_repro[np.isnan(parent_repro)] = 0.0

        # Convert fitness to list of normalized choice probabilities
        parent_repro = np.abs(parent_repro.to_numpy())
        parent_repro_sum = parent_repro.sum()
        parent_repro_mean = parent_repro_sum/len(parent_repro)
        total_repro_mean = pop_fitness/self.pop.shape[0]
        parent_repro_total = parent_repro_mean/(parent_repro_mean + total_repro_mean)
        parent_repro /= np.abs(parent_repro_sum)

        # Get a list containing the indices of all population members
        pop_indices = self.pop.index.values
        parent_indices = self.current_winners.index.values
        for n in range(n_parent_pairs):
            parent_pair = []
            for _ in range(2):
                winner_parent = np.random.choice([True, False], p=[parent_repro_total, 1-parent_repro_total])
                if winner_parent:
                    p_idx = np.random.choice(parent_indices, replace=False, p=parent_repro)
                    parent_pair.append(self.current_winners.loc[p_idx, :])
                else:
                    p_idx = np.random.choice(pop_indices, replace=False)
                    parent_pair.append(self.pop.loc[p_idx, :])
            parents.append(tuple(parent_pair))
        return parents

    def __crossover(self, parent_pairs, n_tries=10):
        """Create a child from each parent pair. Each child gene is uniformly chosen from one of its parents

        If the child already exists in the current population, new genes are chosen, but maximal n_tries times.
        """
        childs = []
        for parents in parent_pairs:
            count = 0
            while count < n_tries:
                child_genes = []
                child_sigma = []
                for g, gene in enumerate(self.initial_gene_pool):
                    choice = np.random.uniform()
                    if choice > 0.5:
                        child_genes.append(parents[0][gene])
                        child_sigma.append(parents[0]['sigma'][g])
                    else:
                        child_genes.append(parents[1][gene])
                        child_sigma.append(parents[1]['sigma'][g])
                already_exists = (self.pop.loc[:, self.gene_names] == child_genes).all(1).any()
                if not already_exists and (child_genes, child_sigma) not in childs:
                    break
                count += 1
            if count < n_tries:
                childs.append((child_genes, child_sigma))
        return childs

    @staticmethod
    def __sample_gene(sampling_func, **kwargs):
        try:
            return sampling_func(**kwargs)
        except TypeError:
            min_val, max_val = kwargs.pop('min'), kwargs.pop('max')
            try:
                vals = list(sampling_func(**kwargs))
                idx = np.argwhere((vals < min_val) + (vals > max_val)).squeeze().tolist()
                while idx:
                    for i in idx:
                        vals.pop(i)
                    kwargs['size'] = len(idx)
                    vals += list(sampling_func(**kwargs))
                    idx = np.argwhere((vals < min_val) + (vals > max_val)).squeeze().tolist()
            except TypeError:
                vals = np.random.uniform(min_val, max_val, kwargs['size'])
            return np.asarray(vals)

    def plot_genes(self, pop_member):
        # Iterate over all genes of the member
        for gene in pop_member.columns.tolist():
            print(f'{gene}:', end=" ")
            data = pop_member[gene].array[0]
            if isinstance(data, list):
                print('[', end="")
                for val in data:
                    print(f'{val}', end=", ")
                print(']', end=" ")
            else:
                print(float(np.round(data, decimals=5)), end=" ")

            # Print borders if available in initial gene pool
            try:
                min = np.round(self.initial_gene_pool[gene]['min'], decimals=2)
                max = np.round(self.initial_gene_pool[gene]['max'], decimals=2)
                print(f' [min: {min}, max: {max}]')
            except KeyError:
                print("")

    def pop_to_grid(self):
        return self.pop.loc[:, self.gene_names]

    def eval_fitness(self, target: list, *argv, **kwargs):
        raise NotImplementedError


class DifferentialEvolutionAlgorithm(GeneticAlgorithmTemplate):

    model_ids = []

    def __init__(self):
        np.random.seed(1234)
        super().__init__()

    def run(self, initial_gene_pool: dict, gene_map: dict, loss_func: callable,
            loss_kwargs: Optional[dict] = None, run_func: Optional[callable] = None,
            run_kwargs: Optional[dict] = None, template: Optional[Union[str, CircuitTemplate]] = None,
            compile_kwargs: Optional[dict] = None, save_dir: Optional[str] = "", verbose: bool = True, **kwargs):
        """Run a genetic algorithm to optimize genes of a population with respect to a given loss function.

        Parameters
        ----------
        initial_gene_pool
            Dictionary containing ranges for each gene to sample from
        gene_map
            Dictionary that provides pointer to model variables for each gene specified in `initial_gene_pool`.
        loss_func
            Function that will be used to evaluate the non-fitness of a certain model candidate.
        loss_kwargs
            Additional keyword arguments that will be passed onto `loss_func`.
        run_func
            Function that will be used to calculate the behavior of a model candidate. The return value of this function
            will be passed onto `loss_func` as first argument. If not provided, you need to specify a `template`,
            the `compile_kwargs` and the `run_func_kwargs`. Those will be used to call `CircuitTemplate.from_yaml`,
            'CircuitIR.compile' and `CircuitIR.run`, respectively.
        run_kwargs
            Additional keyword arguments that will be passed onto `run_func`, if `run_func` is provided. Else, they
            will be passed onto `CircuitIR.run`.
        template
            Either a path to a yaml definition of a `CircuitTemplate` or a `pyrates.frontend.CircuitTemplate` instance.
            Only required, if no `run_func` is provided.
        compile_kwargs
            Additional keyword arguments that will be passed onto `CircuitIR.compile`. Only required if no `run_func`
            is specified.
        verbose
            If true, status updates will be printed.
        kwargs
            Additional keyword arguments that will be passed onto `scipy.optimize.differential_evolution`.

        Returns
        -------
        pandas.DataFrame containing the overall fittest member of all computed populations

        """

        import h5py

        self.initial_gene_pool = initial_gene_pool
        self.num_genes = len(initial_gene_pool)

        # prepare arguments for call to scipy.optimize.differential_evolution
        if verbose:
            print('Starting preparations of evolutionary model optimization.')
        genes = list(initial_gene_pool)
        gene_bounds = [(initial_gene_pool[g]['min'], initial_gene_pool[g]['max']) for g in genes]
        gene_map = [gene_map[g] for g in genes]

        # perform differential evolution optimization
        if verbose:
            print('Starting evolutionary optimization.')
        results = differential_evolution(self.eval_fitness, bounds=gene_bounds,
                                         args=(gene_map, loss_func, loss_kwargs, run_func, template,
                                               compile_kwargs, run_kwargs),
                                         **kwargs)

        # extract results
        final_genes = results.x
        final_loss = results.fun
        optim_sucess = results.success

        # print results
        if verbose:
            print('Finished.')
            print('\n')
            if optim_sucess:
                print('Successfully finished the evolutionary optimization of the model.')
            else:
                print('Failed to finish the evolutionary optimization of the model successfully.')
                print(f'Termination message: \n {results.message}')
            print(f'Final value of the loss function: \n {final_loss}')
            print(f'Final gene set: \n {final_genes}')

        # save results
        self.winner = pd.DataFrame(columns=genes + ['loss'], data=[list(final_genes) + [final_loss]])
        if save_dir:
            path = f'{save_dir}/fittest_candidate.h5'
            print(f"Saving final gene set to: {path}")
            self.winner.to_hdf(path, key='data')

        # remove temporary data
        try:
            from shutil import rmtree
            rmtree('pyrates_build')
        except FileNotFoundError:
            pass

        return self.winner

    def eval_fitness(self, genes: list, gene_map: list, loss_func: callable,
                     loss_kwargs: Optional[dict] = None, run_func: Optional[callable] = None,
                     template: Optional[str] = None, compile_kwargs: Optional[dict] = None,
                     run_kwargs: Optional[dict] = None) -> float:
        """Performs simulation in PyRates of the model defined by `template` and calculates the fitness based on the
        resulting timeseries of that simulation.

        Parameters
        ----------
        genes
            List of model parameters
        gene_map
            List of string-based variable pointers that indicate which gene refers to which variable in the model.
        loss_func
        loss_kwargs
        run_func
            User-specified run function that can be specified instead of a template, to run user-specified routines.
        template
        compile_kwargs
        run_kwargs

        Returns
        -------

        """

        if run_func is None:

            attempts = 1
            while attempts < 100:

                try:

                    # load model template
                    model_id = self.get_unique_id(int(attempts*1e6))
                    model = CircuitTemplate.from_yaml(template).apply(label=f'model_{model_id}')

                    # apply new parameters to model template
                    params, param_map = dict(), dict()
                    for i, (p, key) in enumerate(zip(genes, gene_map)):
                        params[i] = p
                        param_map[i] = key
                    adapt_circuit(model, params=params, param_map=param_map)

                    # compile model into backend
                    model_compiled = model.compile(**compile_kwargs)

                    # define run func
                    run_func = model_compiled.run
                    break

                except (FileExistsError, FileNotFoundError):
                    attempts += 1
                    continue

        results = run_func(**run_kwargs)
        return loss_func(results, **loss_kwargs)

    def get_unique_id(self, upper):
        while True:
            id = np.random.randint(low=0, high=upper)
            if id not in self.model_ids:
                return id


class GSGeneticAlgorithm(GeneticAlgorithmTemplate):
    from scipy.spatial.distance import cdist

    def __init__(self, gs_config, fitness_measure=cdist, **fitness_kwargs):
        super().__init__()

        self.fitness_measure = fitness_measure
        self.fitness_kwargs = fitness_kwargs
        self.gs_config = gs_config

    def eval_fitness(self, target: list, **kwargs):
        param_grid = self.pop.drop(['fitness', 'sigma', 'results'], axis=1)

        results, params = grid_search(circuit_template=self.gs_config['circuit_template'],
                                      param_grid=param_grid,
                                      param_map=self.gs_config['param_map'],
                                      simulation_time=self.gs_config['simulation_time'],
                                      dt=self.gs_config['step_size'],
                                      sampling_step_size=self.gs_config['sampling_step_size'],
                                      permute_grid=False,
                                      inputs=self.gs_config['inputs'],
                                      outputs=self.gs_config['outputs'].copy(),
                                      init_kwargs=self.gs_config['init_kwargs'],
                                      **kwargs
                                      )

        for i, candidate_genes in enumerate(param_grid.values):
            candidate_out = results.loc[:, tuple(candidate_genes)].values.T
            target_reshaped = np.array(target)[None, :]
            dist = self.fitness_measure(candidate_out, target_reshaped, **self.fitness_kwargs)
            self.pop.at[i, 'fitness'] = float(1 / dist)


class CGSGeneticAlgorithm(GeneticAlgorithmTemplate):
    def __init__(self, gs_config, cgs_config, fitness_measure, **fitness_kwargs):
        super().__init__()

        self.gs_config = gs_config
        self.cgs_config = cgs_config
        self.fitness_measure = fitness_measure
        self.fitness_kwargs = fitness_kwargs

        self.cgs = ClusterGridSearch(cgs_config['nodes'], compute_dir=cgs_config['compute_dir'])

    def eval_fitness(self, target: list, **kwargs):

        param_grid = self.pop.drop(['fitness', 'sigma', 'results'], axis=1)

        res_file = self.cgs.run(
            circuit_template=self.gs_config['circuit_template'],
            param_grid=param_grid,
            param_map=self.gs_config['param_map'],
            simulation_time=self.gs_config['simulation_time'],
            dt=self.gs_config['step_size'],
            inputs=self.gs_config['inputs'],
            outputs=self.gs_config['outputs'],
            sampling_step_size=self.gs_config['sampling_step_size'],
            permute=False,
            chunk_size=self.cgs_config['chunk_size'],
            worker_env=self.cgs_config['worker_env'],
            worker_file=self.cgs_config['worker_file'],
            worker_kwargs={'target': target},
            result_concat_axis=0)

        results = pd.read_hdf(res_file, key=f'/Results/fitness')

        for i, candidate_genes in enumerate(param_grid.values):
            self.pop.at[i, 'fitness'] = float(results.loc['fitness', tuple(candidate_genes)])


class ClusterDiffEvAlgorithm(DifferentialEvolutionAlgorithm, ClusterCompute):

    def __init__(self, nodes: list, compute_dir: str, verbose: bool = True):
        super(DifferentialEvolutionAlgorithm).__init__()
        super(ClusterCompute).__init__(nodes, compute_dir=compute_dir, verbose=verbose)

    def run(self, initial_gene_pool: dict, gene_map: dict, loss_func: callable,
            loss_kwargs: Optional[dict] = None, run_func: Optional[callable] = None,
            run_kwargs: Optional[dict] = None, template: Optional[Union[str, CircuitTemplate]] = None,
            compile_kwargs: Optional[dict] = None, verbose: bool = True, time_limit: Optional[float] = None,
            worker_env: Optional[Union[str, list]] = sys.executable, worker_save_dir: Optional[str] = "", **kwargs):

        import h5py

        if verbose:
            sys.stdout = StreamTee(sys.stdout, self.global_logfile)
            sys.stderr = StreamTee(sys.stderr, self.global_logfile)
        else:
            sys.stdout = open(self.global_logfile, 'a')
            sys.stderr = StreamTee(sys.stderr, self.global_logfile)

        t_total = t.time()

        # Add local environment to each worker
        ######################################

        for i, client in enumerate(self.clients):
            if isinstance(worker_env, list):
                client['worker_env'] = worker_env[i]
            else:
                client['worker_env'] = worker_env

        # Create global result file
        ############################

        if verbose:
            print("***Creating global results file***")
        t0 = t.time()

        # Create result directory and result file for current parameter grid
        global_res_file = f'{self.compute_dir}/optimization_result.h5'

        # Write gene pool and config information to global result file
        with h5py.File(global_res_file, 'a') as file:
            for key, value in initial_gene_pool.items():
                file.create_dataset(f'/GenePool/Keys/{key}', data=value)
        # TODO: add global configuration details to file
        if verbose:
            print(f'Done. Elapsed time: {t.time() - t0:.3f} seconds \n\n')

        # Create keyword dictionary for threads
        #######################################
        thread_kwargs = {
            "worker_env": worker_env,
            "initial_gene_pool": initial_gene_pool,
            "gene_map": gene_map,
            "loss_func": loss_func,
            "loss_kwargs": loss_kwargs,
            "run_func": run_func,
            "run_kwargs": run_kwargs,
            "compile_kwargs": compile_kwargs,
            "template": template,
            "save_dir": worker_save_dir
        }

        # Start cluster computation
        ###########################

        if verbose:
            print("***Starting cluster computation***")

        # Spawn threads to control each node connection and start computation
        threads = [self.spawn_thread(client, thread_kwargs, timeout=time_limit) for client in self.clients]
        # Wait for all threads to finish
        for t_ in threads:
            t_.join()

        print("")
        print(f'Cluster computation finished. Elapsed time: {t.time() - t_total:.3f} seconds')

        # Write local results to global result file
        ###########################################

        print(f'***WRITING RESULTS TO GLOBAL RESULT FILE***')
        t0 = t.time()

        # Get sorted list of temporary result files to iterate through
        temp_res_files = glob.glob(f'{self.compute_dir}/*_temp*')
        temp_res_files.sort()

        # Read number of different circuit outputs and prepare lists to concatenate results
        res_dict = {}
        try:
            with pd.HDFStore(temp_res_files[0], "r") as store:
                for key in store.keys():
                    res_dict[key] = []
        except IndexError:
            print("No result file created. Check local log files for worker script errors.")
            return ""

        # Concatenate results from each temporary result file
        for file in temp_res_files:
            with pd.HDFStore(file, "r") as store:
                for idx, key in enumerate(store.keys()):
                    res_dict[key].append(store[key])

        # Create DataFrame for each output variable and write to global result file
        with pd.HDFStore(global_res_file, "a") as store:
            for key, value in res_dict.items():
                if key != '/result_map' and len(value) > 0:
                    df = pd.concat(value, axis=kwargs.pop('result_concat_axis', 1))
                    store.put(key=f'/Results{key}', value=df)
            result_map = pd.concat(res_dict['/result_map'], axis=0)
            store.put(key=f'/Results/result_map', value=result_map)

        # Delete temporary local result files
        for file in temp_res_files:
            os.remove(file)

        print(f'Elapsed time: {t.time() - t0:.3f} seconds')
        print(f'Find results in: {self.compute_dir}/optimization_results.h5')
        print("")

        return global_res_file

    def thread_master(self, client, thread_kwargs: dict, timeout: float):
        """

        Parameters
        ----------
        client
        thread_kwargs
        timeout

        Returns
        -------

        """

        from threading import currentThread

        # This lock ensures that parameter chunks are fetched by workers in the same order as they are defined
        # in the node list
        self.lock.acquire()

        thread_name = currentThread().getName()

        # Get client information
        pm_client = client["paramiko_client"]
        logfile = client["logfile"]
        worker_env = client["worker_env"]

        # Get keyword arguments
        config_file = thread_kwargs["config_file"]
        worker_file = thread_kwargs["worker_file"]

        # Prepare worker command
        command = f'{worker_env} {worker_file}'

        # Create build_dir for grid_search numpy backend
        local_build_dir = f'{self.compute_dir}/{thread_name}'
        os.makedirs(local_build_dir, exist_ok=True)

        # Start scheduler
        #################

        # Disable thread switching
        self.lock.acquire()

        # Execute worker script on the remote host
        print(f'[T]\'{thread_name}\': Starting remote computation...')

        channel = pm_client.get_transport().open_session()
        channel.settimeout(timeout)

        # Execute the given command
        channel.get_pty()
        channel.exec_command(command +
                             f' --config_file={config_file}'
                             f' --local_res_file={local_build_dir}'
                             f' --build_dir={local_build_dir}'
                             f' &>{logfile}',  # redirect and append stdout
                             )

        # Enable thread switching
        self.lock.release()

        # Release the second lock if its still acquired from the very beginning
        try:
            self.lock.release()
        except RuntimeError:
            pass

        # Wait for remote computation exit status
        # (independent of success or failure)
        #########################################

        nbytes = 1024
        t1 = t.time()
        while not channel.exit_status_ready():
            if channel.recv_ready():
                data = channel.recv(nbytes)
                while data:
                    data = channel.recv(nbytes)
            if channel.recv_stderr_ready():
                error_buff = channel.recv_stderr(nbytes)
                while error_buff:
                    error_buff = channel.recv_stderr(nbytes)
            t2 = t.time()
            if t2 - t1 > timeout:
                channel.close()
                break
            t.sleep(1.0)
        channel.close()
        exit_status = channel.recv_exit_status()

        return exit_status


def plot_results_2d(p1: str, p2: str, fname_identifier: str, fname_type: str = '.h5', fitness_measure: str = 'fitness',
                    cbar_kwargs: dict = None, **kwargs):
    """

    Parameters
    ----------
    p1
    p2
    fname_identifier
    fname_type
    fitness_measure
    cbar_kwargs
    kwargs

    Returns
    -------

    """
    import glob
    import os
    from pandas import read_hdf, DataFrame
    from seaborn import jointplot
    from matplotlib.pyplot import colorbar, subplots_adjust

    # retrieve data from files
    if '/' not in fname_identifier:
        fname_identifier = f"{os.curdir}/{fname_identifier}"
    files = glob.glob(f"{fname_identifier}*{fname_type}")
    p1_col, p2_col, fitness = [], [], []
    for f in files:
        data = read_hdf(f, key='data')
        p1_col += list(data.loc[:, p1])
        p2_col += list(data.loc[:, p2])
        fitness += list(data.loc[:, fitness_measure])

    # plot data
    grid = jointplot(p1_col, p2_col, C=fitness, **kwargs)

    # add subplot
    pos = cbar_kwargs.pop('pos', [.85, .25, .05, .4])
    left = cbar_kwargs.pop('left', 0.2)
    right = cbar_kwargs.pop('right', 0.8)
    top = cbar_kwargs.pop('top', 0.8)
    bottom = cbar_kwargs.pop('bottom, 0.2')
    subplots_adjust(left=left, right=right, top=top, bottom=bottom)
    cbar_ax = grid.fig.add_axes(pos)  # x, y, width, height
    colorbar(cax=cbar_ax)

    return grid


# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

"""Wraps tensorflow such that it's low-level functions can be used by PyRates to create and simulate a compute graph.
"""

# external imports
from typing import Optional, Dict, Callable, List, Any, Union
import os
import sys
from shutil import rmtree
import numpy as np
from numpy import f2py

# pyrates internal imports
from .numpy_backend import NumpyBackend, PyRatesAssignOp, PyRatesIndexOp, PyRatesOp, CodeGen, extract_lhs_var

# meta infos
__author__ = "Richard Gast"
__status__ = "development"

module_counter = 0


class FortranOp(PyRatesOp):

    def __init__(self, op: str, short_name: str, name, *args, **kwargs) -> None:
        self.build_dir = kwargs.pop('build_dir', '')
        super().__init__(op, short_name, name, *args, **kwargs)

    def _generate_func(self):
        return generate_func(self)

    @classmethod
    def _process_args(cls, args, results, constants_to_num=False):
        return super()._process_args(args, results, constants_to_num=constants_to_num)


class FortranIndexOp(PyRatesIndexOp):

    def __init__(self, op: str, short_name: str, name, *args, **kwargs) -> None:
        self.build_dir = kwargs.pop('build_dir', '')
        super().__init__(op, short_name, name, *args, **kwargs)

    def _generate_func(self):
        return generate_func(self)


class FortranAssignOp(PyRatesAssignOp):

    def __init__(self, op: str, short_name: str, name, *args, **kwargs) -> None:
        self.build_dir = kwargs.pop('build_dir', '')
        super().__init__(op, short_name, name, *args, **kwargs)

    def _generate_func(self, return_key='y_delta'):
        try:
            idx = self._op_dict['arg_names'].index(return_key)
        except ValueError:
            return_key = extract_lhs_var(self._op_dict['value'])
            idx = self._op_dict['arg_names'].index(return_key)
        ndim = self._op_dict['args'][idx].shape
        return generate_func(self, return_key=return_key, omit_assign=True, return_dim=ndim, return_intent='out')

    def numpy(self):
        result = self._callable(*self.args[1:])
        self._check_numerics(result, self.name)
        return result


class FortranBackend(NumpyBackend):

    idx_l, idx_r = "(", ")"
    idx_start = 1

    def __init__(self,
                 ops: Optional[Dict[str, str]] = None,
                 dtypes: Optional[Dict[str, object]] = None,
                 name: str = 'net_0',
                 float_default_type: str = 'float32',
                 imports: Optional[List[str]] = None,
                 build_dir: Optional[str] = None,
                 auto_compat: bool = False
                 ) -> None:
        """Instantiates numpy backend, i.e. a compute graph with numpy operations.
        """

        # define operations and datatypes of the backend
        ################################################

        # base math operations
        ops_f = {"+": {'name': "fortran_add", 'call': "+"},
                 "-": {'name': "fortran_subtract", 'call': "-"},
                 "*": {'name': "fortran_multiply", 'call': "*"},
                 "/": {'name': "fortran_divide", 'call': "/"},
                 "%": {'name': "fortran_modulo", 'call': "MODULO"},
                 "^": {'name': "fortran_power", 'call': "**"},
                 "**": {'name': "fortran_power_float", 'call': "**"},
                 "@": {'name': "fortrandot", 'call': ""},
                 ".T": {'name': "fortrantranspose", 'call': ""},
                 ".I": {'name': "fortraninvert", 'call': ""},
                 ">": {'name': "fortrangreater", 'call': ">"},
                 "<": {'name': "fortranless", 'call': "<"},
                 "==": {'name': "fortranequal", 'call': "=="},
                 "!=": {'name': "fortran_not_equal", 'call': "!="},
                 ">=": {'name': "fortran_greater_equal", 'call': ">="},
                 "<=": {'name': "fortran_less_equal", 'call': "<="},
                 "=": {'name': "assign", 'call': "="},
                 "+=": {'name': "assign_add", 'call': ""},
                 "-=": {'name': "assign_subtract", 'call': ""},
                 "*=": {'name': "assign_multiply", 'call': ""},
                 "/=": {'name': "assign_divide", 'call': ""},
                 "neg": {'name': "negative", 'call': "-"},
                 "sin": {'name': "fortran_sin", 'call': "SIN"},
                 "cos": {'name': "fortran_cos", 'call': "COS"},
                 "tan": {'name': "fortran_tan", 'call': "TAN"},
                 "atan": {'name': "fortran_atan", 'call': "ATAN"},
                 "abs": {'name': "fortran_abs", 'call': "ABS"},
                 "sqrt": {'name': "fortran_sqrt", 'call': "SQRT"},
                 "sq": {'name': "fortran_square", 'call': ""},
                 "exp": {'name': "fortran_exp", 'call': "EXP"},
                 "max": {'name': "fortran_max", 'call': "MAX"},
                 "min": {'name': "fortran_min", 'call': "MIN"},
                 "argmax": {'name': "fortran_transpose", 'call': "ARGMAX"},
                 "argmin": {'name': "fortran_argmin", 'call': "ARGMIN"},
                 "round": {'name': "fortran_round", 'call': ""},
                 "sum": {'name': "fortran_sum", 'call': "SUM"},
                 "mean": {'name': "fortran_mean", 'call': ""},
                 "concat": {'name': "fortran_concatenate", 'call': ""},
                 "reshape": {'name': "fortran_reshape", 'call': ""},
                 "append": {'name': "fortran_append", 'call': ""},
                 "shape": {'name': "fortran_shape", 'call': ""},
                 "dtype": {'name': "fortran_dtype", 'call': ""},
                 'squeeze': {'name': "fortran_squeeze", 'call': ""},
                 'expand': {'name': 'fortran_expand', 'call': ""},
                 "roll": {'name': "fortran_roll", 'call': ""},
                 "cast": {'name': "fortran_cast", 'call': ""},
                 "randn": {'name': "fortran_randn", 'call': ""},
                 "ones": {'name': "fortran_ones", 'call': ""},
                 "zeros": {'name': "fortran_zeros", 'call': ""},
                 "range": {'name': "fortran_arange", 'call': ""},
                 "softmax": {'name': "pyrates_softmax", 'call': ""},
                 "sigmoid": {'name': "pyrates_sigmoid", 'call': ""},
                 "tanh": {'name': "fortran_tanh", 'call': "TANH"},
                 "index": {'name': "pyrates_index", 'call': "pyrates_index"},
                 "mask": {'name': "pyrates_mask", 'call': ""},
                 "group": {'name': "pyrates_group", 'call': ""},
                 "asarray": {'name': "fortran_asarray", 'call': ""},
                 "no_op": {'name': "pyrates_identity", 'call': "no_op"},
                 "interpolate": {'name': "pyrates_interpolate", 'call': ""},
                 "interpolate_1d": {'name': "pyrates_interpolate_1d", 'call': ""},
                 "interpolate_nd": {'name': "pyrates_interpolate_nd", 'call': ""},
                 }
        if ops:
            ops_f.update(ops)
        self.pyauto_compat = auto_compat
        super().__init__(ops=ops_f, dtypes=dtypes, name=name, float_default_type=float_default_type,
                         imports=imports, build_dir=build_dir)
        self._imports = []
        self.npar = 0
        self.ndim = 0
        self._auto_files_generated = False

    def compile(self, build_dir: Optional[str] = None, decorator: Optional[Callable] = None, **kwargs) -> tuple:
        """Compile the graph layers/operations. Creates python files containing the functions in each layer.

        Parameters
        ----------
        build_dir
            Directory in which to create the file structure for the simulation.
        decorator
            Decorator function that should be applied to the right-hand side evaluation function.
        kwargs
            decorator keyword arguments

        Returns
        -------
        tuple
            Contains tuples of layer run functions and their respective arguments.

        """

        # preparations
        ##############

        # remove empty layers and operators
        new_layer_idx = 0
        for layer_idx, layer in enumerate(self.layers.copy()):
            for op in layer.copy():
                if op is None:
                    layer.pop(layer.index(op))
            if len(layer) == 0:
                self.layers.pop(new_layer_idx)
            else:
                new_layer_idx += 1

        # remove previously imported rhs_funcs from system
        if 'rhs_func' in sys.modules:
            del sys.modules['rhs_func']

        # collect state variable and parameter vectors
        state_vars, params, var_map = self._process_vars()

        # update state variable getter operations to include the full state variable vector as first argument
        y = self.get_var('y')
        for key, (vtype, idx) in var_map.items():
            var = self.get_var(key)
            if vtype == 'state_var' and var.short_name != 'y':
                var.args[0] = y
                var.build_op(var.args)

        # create rhs evaluation function
        ################################

        # set up file header
        func_gen = FortranGen()
        for import_line in self._imports:
            func_gen.add_code_line(import_line)
            func_gen.add_linebreak()
        func_gen.add_linebreak()

        # define function head
        func_gen.add_indent()
        if self.pyauto_compat:
            func_gen.add_code_line("subroutine func(ndim,y,icp,args,ijac,y_delta,dfdu,dfdp)")
            func_gen.add_linebreak()
            func_gen.add_code_line("implicit none")
            func_gen.add_linebreak()
            func_gen.add_code_line("integer, intent(in) :: ndim, icp(*), ijac")
            func_gen.add_linebreak()
            func_gen.add_code_line("double precision, intent(in) :: y(ndim), args(*)")
            func_gen.add_linebreak()
            func_gen.add_code_line("double precision, intent(out) :: y_delta(ndim)")
            func_gen.add_linebreak()
            func_gen.add_code_line("double precision, intent(inout) :: dfdu(ndim,ndim), dfdp(ndim,*)")
            func_gen.add_linebreak()
        else:
            func_gen.add_code_line("subroutine func(ndim,t,y,args,y_delta)")
            func_gen.add_linebreak()
            func_gen.add_code_line("implicit none")
            func_gen.add_linebreak()
            func_gen.add_code_line("integer, intent(in) :: ndim")
            func_gen.add_linebreak()
            func_gen.add_code_line("double precision, intent(in) :: t, y(ndim)")
            func_gen.add_linebreak()
            func_gen.add_code_line("double precision, intent(in) :: args(*)")
            func_gen.add_linebreak()
            func_gen.add_code_line("double precision, intent(out) :: y_delta(ndim)")
            func_gen.add_linebreak()

        # declare variable types
        func_gen.add_code_line("double precision ")
        for key in var_map:
            var = self.get_var(key)
            if "float" in str(var.dtype) and var.short_name != 'y_delta':
                func_gen.add_code_line(f"{var.short_name},")
        if "," in func_gen.code[-1]:
            func_gen.code[-1] = func_gen.code[-1][:-1]
        else:
            func_gen.code.pop(-1)
        func_gen.add_linebreak()
        func_gen.add_code_line("integer ")
        for key in var_map:
            var = self.get_var(key)
            if "int" in str(var.dtype):
                func_gen.add_code_line(f"{var.short_name},")
        if "," in func_gen.code[-1]:
            func_gen.code[-1] = func_gen.code[-1][:-1]
        else:
            func_gen.code.pop(-1)
        func_gen.add_linebreak()

        # declare constants
        args = [None for _ in range(len(params))]
        func_gen.add_code_line("! declare constants")
        func_gen.add_linebreak()
        updates, indices = [], []
        i = 0
        for key, (vtype, idx) in var_map.items():
            if vtype == 'constant':
                var = params[idx-self.idx_start][1]
                if var.short_name != 'y_delta':
                    func_gen.add_code_line(f"{var.short_name} = args({idx})")
                    func_gen.add_linebreak()
                    args[idx-1] = var
                    updates.append(f"{var.short_name}")
                    indices.append(i)
                    i += 1
        func_gen.add_linebreak()

        # extract state variables from input vector y
        func_gen.add_code_line("! extract state variables from input vector")
        func_gen.add_linebreak()
        for key, (vtype, idx) in var_map.items():
            var = self.get_var(key)
            if vtype == 'state_var':
                func_gen.add_code_line(f"{var.short_name} = {var.value}")
                func_gen.add_linebreak()
        func_gen.add_linebreak()

        # add equations
        func_gen.add_code_line("! calculate right-hand side update of equation system")
        func_gen.add_linebreak()
        arg_updates = []
        for i, layer in enumerate(self.layers):
            for j, op in enumerate(layer):
                lhs = op.value.split("=")[0]
                lhs = lhs.replace(" ", "")
                find_arg = [arg == lhs for arg in updates]
                if any(find_arg):
                    idx = find_arg.index(True)
                    arg_updates.append((updates[idx], indices[idx]))
                func_gen.add_code_line(op.value)
                func_gen.add_linebreak()
        func_gen.add_linebreak()

        # update parameters where necessary
        # func_gen.add_code_line("! update system parameters")
        # func_gen.add_linebreak()
        # for upd, idx in arg_updates:
        #     update_str = f"args{self.idx_l}{idx}{self.idx_r} = {upd}"
        #     if f"    {update_str}" not in func_gen.code:
        #         func_gen.add_code_line(update_str)
        #         func_gen.add_linebreak()

        # end function
        func_gen.add_code_line(f"end subroutine func")
        func_gen.add_linebreak()
        func_gen.remove_indent()

        # save rhs function to file
        fname = f'{self._build_dir}/rhs_func'
        f2py.compile(func_gen.generate(), modulename='rhs_func', extension='.f', source_fn=f'{fname}.f', verbose=False)

        # create additional subroutines in pyauto compatibility mode
        gen_def = kwargs.pop('generate_auto_def', True)
        if self.pyauto_compat and gen_def:
            self.generate_auto_def(self._build_dir)

        # import function from file
        fname_import = fname.replace('/', '.')
        exec(f"from rhs_func import rhs_eval", globals())
        rhs_eval = globals().pop('func')

        # apply function decorator
        if decorator:
            rhs_eval = decorator(rhs_eval, **kwargs)

        return rhs_eval, args, state_vars, var_map

    def generate_auto_def(self, directory):
        """

        Parameters
        ----------
        directory

        Returns
        -------

        """

        if not self.pyauto_compat:
            raise ValueError('This method can only be called in pyauto compatible mode. Please set `pyauto_compat` to '
                             'True upon calling the `CircuitIR.compile` method.')

        # read file
        ###########

        try:

            # read file from excisting system compilation
            if not directory:
                directory = self._build_dir
            fn = f"{directory}/rhs_func.f" if "rhs_func.f" not in directory else directory
            with open(fn, 'rt') as f:
                func_str = f.read()

        except FileNotFoundError:

            # compile system and then read the build files
            self.compile(build_dir=directory, generate_auto_def=False)
            fn = f"{self._build_dir}/rhs_func.f"
            with open(fn, 'r') as f:
                func_str = f.read()

        end_str = "end subroutine func\n"
        idx = func_str.index(end_str)
        func_str = func_str[:idx+len(end_str)]

        # generate additional subroutines
        #################################

        func_gen = FortranGen()

        # generate subroutine header
        func_gen.add_linebreak()
        func_gen.add_indent()
        func_gen.add_code_line("subroutine stpnt(ndim, y, args, t)")
        func_gen.add_linebreak()
        func_gen.add_code_line("implicit None")
        func_gen.add_linebreak()
        func_gen.add_code_line("integer, intent(in) :: ndim")
        func_gen.add_linebreak()
        func_gen.add_code_line("double precision, intent(inout) :: y(ndim), args(*)")
        func_gen.add_linebreak()
        func_gen.add_code_line("double precision, intent(in) :: T")
        func_gen.add_linebreak()

        # declare variable types
        func_gen.add_code_line("double precision ")
        for key in self.vars:
            var = self.get_var(key)
            name = var.short_name
            if "float" in str(var.dtype) and name != 'y_delta' and name != 'y' and name != 't':
                func_gen.add_code_line(f"{var.short_name},")
        if "," in func_gen.code[-1]:
            func_gen.code[-1] = func_gen.code[-1][:-1]
        else:
            func_gen.code.pop(-1)
        func_gen.add_linebreak()
        func_gen.add_code_line("integer ")
        for key in self.vars:
            var = self.get_var(key)
            if "int" in str(var.dtype):
                func_gen.add_code_line(f"{var.short_name},")
        if "," in func_gen.code[-1]:
            func_gen.code[-1] = func_gen.code[-1][:-1]
        else:
            func_gen.code.pop(-1)
        func_gen.add_linebreak()

        _, params, var_map = self._process_vars()

        # define parameter values
        func_gen.add_linebreak()
        for key, (vtype, idx) in var_map.items():
            if vtype == 'constant':
                var = params[idx-self.idx_start][1]
                if var.short_name != 'y_delta' and var.short_name != 'y':
                    func_gen.add_code_line(f"{var.short_name} = {var}")
                    func_gen.add_linebreak()
        func_gen.add_linebreak()

        # define initial state
        func_gen.add_linebreak()
        npar = 0
        for key, (vtype, idx) in var_map.items():
            if vtype == 'constant':
                var = params[idx-self.idx_start][1]
                if var.short_name != 'y_delta' and var.short_name != 'y':
                    func_gen.add_code_line(f"args({idx}) = {var.short_name}")
                    func_gen.add_linebreak()
                    if idx > npar:
                        npar = idx
        func_gen.add_linebreak()

        func_gen.add_linebreak()
        for key, (vtype, idx) in var_map.items():
            var = self.get_var(key)
            if vtype == 'state_var':
                func_gen.add_code_line(f"{var.value} = {var.numpy()}")
                func_gen.add_linebreak()
        func_gen.add_linebreak()

        # end subroutine
        func_gen.add_linebreak()
        func_gen.add_code_line("end subroutine stpnt")
        func_gen.add_linebreak()

        # add dummy subroutines
        for routine in ['bcnd', 'icnd', 'fopt', 'pvls']:
            func_gen.add_linebreak()
            func_gen.add_code_line(f"subroutine {routine}")
            func_gen.add_linebreak()
            func_gen.add_code_line(f"end subroutine {routine}")
            func_gen.add_linebreak()
        func_gen.add_linebreak()
        func_gen.remove_indent()

        func_combined = f"{func_str} \n {func_gen.generate()}"
        f2py.compile(func_combined, source_fn=fn, modulename='rhs_func', extension='.f', verbose=False)

        self.npar = npar
        self.ndim = self.get_var('y').shape[0]

        # generate constants file
        #########################

        # declare auto constants and their values
        auto_constants = {'NDIM': self.ndim, 'NPAR': self.npar, 'IPS': -2, 'ILP': 0, 'ICP': [14], 'NTST': 1, 'NCOL': 4,
                          'IAD': 3, 'ISP': 0, 'ISW': 1, 'IPLT': 0, 'NBC': 0, 'NINT': 0, 'NMX': 10000, 'NPR': 100,
                          'MXBF': 10, 'IID': 2, 'ITMX': 8, 'ITNW': 5, 'NWTN': 3, 'JAC': 0, 'EPSL': 1e-7, 'EPSU': 1e-7,
                          'EPSS': 1e-5, 'IRS': 0, 'DS': 1e-4, 'DSMIN': 1e-8, 'DSMAX': 1e-2, 'IADS': 1, 'THL': {},
                          'THU': {}, 'UZR': {}, 'STOP': {}}

        # write auto constants to string
        cgen = FortranGen()
        for key, val in auto_constants.items():
            cgen.add_code_line(f"{key} = {val}")
            cgen.add_linebreak()

        # write auto constants to file
        try:
            with open(f'{directory}/c.ivp', 'wt') as cfile:
                cfile.write(cgen.generate())
        except FileNotFoundError:
            with open(f'{directory}/c.ivp', 'xt') as cfile:
                cfile.write(cgen.generate())

        self._auto_files_generated = True

        return fn

    def to_pyauto(self, directory=None, generate_auto_def=True):
        """

        Parameters
        ----------
        directory
        generate_auto_def

        Returns
        -------

        """
        from pyrates.utility.pyauto import PyAuto
        directory = directory if directory else self._build_dir
        if generate_auto_def:
            self.generate_auto_def(directory)
        return PyAuto(directory)

    def clear(self) -> None:
        """Removes all layers, variables and operations from graph. Deletes build directory.
        """

        super().clear()
        for f in [f for f in os.listdir(os.getcwd()) if "cpython" in f and ("rhs_func" in f or "pyrates_func" in f)]:
            os.remove(f)

    def _solve(self, rhs_func, func_args, T, dt, dts, t, solver, output_indices, **kwargs):
        """

        Parameters
        ----------
        rhs_func
        func_args
        state_vars
        T
        dt
        dts
        t
        solver
        output_indices
        kwargs

        Returns
        -------

        """

        if self.pyauto_compat:

            from pyrates.utility.pyauto import PyAuto
            pyauto = PyAuto(auto_dir=self._build_dir)
            dsmin = dt*1e-2
            auto_defs = {'DSMIN': dsmin, 'DSMAX': dt*1e2, 'NMX': int(T/dsmin), 'NPR': int(dts/dt)}
            for key, val in auto_defs.items():
                if key not in kwargs:
                    kwargs[key] = val
            pyauto.run(e='rhs_func', c='ivp', DS=dt, name='t', UZR={14: T}, STOP={'UZ1'}, **kwargs)

            extract = [f'U({i+1})' for i in range(self.ndim)]
            out_vars = [f'U({i[0]+self.idx_start if type(i) is list else i+self.idx_start})' for i in output_indices]
            extract.append('PAR(14)')

            results = pyauto.extract(keys=extract, cont='t')
            times = results.pop('PAR(14)')

            state_vars = self.get_var('y')
            for key, val in results.items():
                start, stop = key.index('('), key.index(')')
                idx = int(key[start+1:stop])
                state_vars[idx-self.idx_start] = val[-1]

            return times, [results[v] for v in out_vars]

        return super()._solve(rhs_func, func_args, T, dt, dts, t, solver, output_indices, **kwargs)

    def _create_op(self, op, name, *args):
        if not self.ops[op]['call']:
            raise NotImplementedError(f"The operator `{op}` is not implemented for this backend ({self.name}). "
                                      f"Please consider passing the required operation to the backend initialization "
                                      f"or choose another backend.")
        if op in ["=", "+=", "-=", "*=", "/="]:
            if len(args) > 2 and hasattr(args[2], 'shape') and len(args[2].shape) > 1 and \
                    'bool' not in str(type(args[2])):
                args_tmp = list(args)
                if not hasattr(args_tmp[2], 'short_name'):
                    idx = self.add_var(vtype='state_var', name='idx', value=args_tmp[2])
                else:
                    idx = args_tmp[2]
                var, upd, idx = self._process_update_args_old(args[0], args[1], idx)
                idx_str = ",".join([f"{idx.short_name}[:,{i}]" for i in range(idx.shape[1])])
                args = (var, upd, idx_str, idx)
            return FortranAssignOp(self.ops[op]['call'], self.ops[op]['name'], name, *args, idx_l=self.idx_l,
                                   idx_r=self.idx_r, build_dir=self._build_dir)
        elif op is "index":
            return FortranIndexOp(self.ops[op]['call'], self.ops[op]['name'], name, *args, idx_l=self.idx_l,
                                  idx_r=self.idx_r, build_dir=self._build_dir)
        else:
            if op is "cast":
                args = list(args)
                for dtype in self.dtypes:
                    if dtype in str(args[1]):
                        args[1] = f"np.{dtype}"
                        break
                args = tuple(args)
            return FortranOp(self.ops[op]['call'], self.ops[op]['name'], name, *args, build_dir=self._build_dir)

    def _process_vars(self):
        """

        Returns
        -------

        """
        state_vars, constants, var_map = [], [], {}
        s_idx, c_idx, s_len = 0, 0, 0
        for key, var in self.vars.items():
            key, state_var = self._is_state_var(key)
            if state_var:
                state_vars.append((key, var))
                var_map[key] = ('state_var', (s_idx, s_len))
                s_idx += 1
                s_len += 1
            elif var.vtype == 'constant' or (var.short_name not in self.lhs_vars and key != 'y'):
                if c_idx == 10:
                    for _ in range(4):
                        constants.append(())
                    c_idx = 14
                constants.append((key, var))
                var_map[key] = ('constant', c_idx+self.idx_start)
                c_idx += 1
        return state_vars, constants, var_map

    @staticmethod
    def _compare_shapes(op1: Any, op2: Any, index=False) -> bool:
        """Checks whether the shapes of op1 and op2 are compatible with each other.

        Parameters
        ----------
        op1
            First operator.
        op2
            Second operator.

        Returns
        -------
        bool
            If true, the shapes of op1 and op2 are compatible.

        """

        if hasattr(op1, 'shape') and hasattr(op2, 'shape'):
            if op1.shape == op2.shape:
                return True
            elif len(op1.shape) > 1 and len(op2.shape) > 1:
                return True
            elif len(op1.shape) == 0 and len(op2.shape) == 0:
                return True
            else:
                return False
        elif hasattr(op1, 'shape'):
            if sum(op1.shape) > 0:
                return False
            else:
                return True
        else:
            return True


class FortranGen(CodeGen):

    def add_code_line(self, code_str):
        """Add code line string to code.
        """
        if self.code:
            idx = -1
            while self.code[idx] != '\n' and abs(idx) < len(self.code):
                idx -= 1
            code_line = ''.join([self.code.pop(i) for i in range(idx+1, 0)]) + code_str
        else:
            code_line = code_str
        if "&" not in code_line:
            code_line = code_line.replace('\t', '')
            code_line = "\t" * self.lvl + code_line
        n = 60
        if len(code_line) > n:
            idx = self._find_first_op(code_line, start=0, stop=n)
            self.code.append(f"{code_line[0:idx]}")
            while idx < len(code_line):
                self.add_linebreak()
                idx_new = self._find_first_op(code_line, start=idx, stop=idx+n)
                self.code.append(f"     & {code_line[idx:idx+idx_new]}")
                idx += idx_new
        else:
            self.code.append(code_line)

    @staticmethod
    def _find_first_op(code, start, stop):
        if stop < len(code):
            code_tmp = code[start:stop]
            ops = ["+", "-", "*", "/", "**", "^", "%", "<", ">", "==", "!=", "<=", ">="]
            indices = [code_tmp.index(op) for op in ops if op in code_tmp]
            if indices and max(indices) > 0:
                return max(indices)
            idx = start
            for break_sign in [',', ')', ' ']:
                if break_sign in code_tmp:
                    idx_tmp = len(code_tmp) - code_tmp[::-1].index(break_sign)
                    if len(code_tmp)-idx_tmp < len(code_tmp)-idx:
                        idx = idx_tmp
            return idx
        return stop


def generate_func(self, return_key='f', omit_assign=False, return_dim=None, return_intent='out'):
    """Generates a function from operator value and arguments"""

    global module_counter

    # function head
    func_dict = {}
    func = FortranGen()
    func.add_linebreak()
    func.add_indent()
    fname = self.short_name.lower()
    func.add_code_line(f"subroutine {fname}({return_key}")
    for arg in self._op_dict['arg_names']:
        if arg != return_key:
            func.add_code_line(f",{arg}")
    func.add_code_line(")")
    func.add_linebreak()

    # argument type definition
    for arg, name in zip(self._op_dict['args'], self._op_dict['arg_names']):
        if name != return_key:
            dtype = "integer" if "int" in str(arg.vtype) else "double precision"
            dim = f"dimension({','.join([str(s) for s in arg.shape])}), " if arg.shape else ""
            func.add_code_line(f"{dtype}, {dim}intent(in) :: {name}")
            func.add_linebreak()
    out_dim = f"({','.join([str(s) for s in return_dim])})" if return_dim else ""
    func.add_code_line(f"double precision, intent({return_intent}) :: {return_key}{out_dim}")
    func.add_linebreak()

    func.add_code_line(f"{self._op_dict['value']}" if omit_assign else f"{return_key} = {self._op_dict['value']}")
    func.add_linebreak()
    func.add_code_line("end")
    func.add_linebreak()
    func.remove_indent()
    module_counter += 1
    fn = f"pyrates_func_{module_counter}"
    f2py.compile(func.generate(), modulename=fn, extension=".f", verbose=False,
                 source_fn=f"{self.build_dir}/{fn}.f" if self.build_dir else f"{fn}.f")
    exec(f"from pyrates_func_{module_counter} import {fname}", globals())
    func_dict[self.short_name] = globals().pop(fname)
    return func_dict
